package comps
{
	import flash.display.DisplayObject;
	import flash.events.AsyncErrorEvent;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.events.NetStatusEvent;
	import flash.media.SoundTransform;
	import flash.media.Video;
	import flash.net.NetConnection;
	import flash.net.NetStream;
	
	import mx.controls.Button;
	import mx.controls.HSlider;
	import mx.core.EdgeMetrics;
	import mx.core.IFlexDisplayObject;
	import mx.core.IRectangularBorder;
	import mx.core.IUIComponent;
	import mx.core.UIComponent;
	import mx.events.FlexEvent;
	import mx.events.SliderEvent;
	import mx.events.VideoEvent;
	import mx.skins.halo.HaloBorder;
	import mx.styles.ISimpleStyleClient;
	
	include "BorderStyles.as" 
	
	/**
	 *  Dispatched when the video is loaded and ready to play.
	 *
	 *  @eventType mx.events.VideoEvent.READY
	 */
	[Event(name="ready", type="mx.events.VideoEvent")]
	
	/**
	 *  The placement of the control buttons (play, stop, volume slider, etc). 
	 *  They can either be banked in a horizontal row below or above the 
	 *  video, or banked vertically to the left or right of the video. 
	 * 
	 *  @default "bottom"
	 *  
	 */
	[Style(name="controlBarPlacement", type="String", inherit="no")]
	
	//By extending UIComponent, I inherit all of the lifecycle methods, 
	//events and properties. Also, since UIComponent extends from 
	//EventDispatcher, VideoWidget inherits the ability to listen and 
	//dispatch events. 
	public class VideoWidget extends UIComponent
	{
		
		//--------------------------------------------------------------------------
    	//
    	//  Constructor - Remember: 0 required arguments. 
    	//
    	//--------------------------------------------------------------------------		
		public function VideoWidget()
		{
			//Remember: Always call super(). 
			super();
			
			//Once my component is fully created, I need to know whether start playing
			//the video based on the autoPlay property. By listening to the 
			//creationComplete event, I can handle the autoPlay functionality.   
			addEventListener(FlexEvent.CREATION_COMPLETE, creationCompleteHandler);
		}
		
		//--------------------------------------------------------------------------
    	//
    	//  Variables 
    	//
    	//--------------------------------------------------------------------------
		
		//The Flash Player video instance that lets you stream video without 
		//embedding the video in your SWF file. 
		private var video:Video;
		
		//The NetConnection which creates a connection between the Flash Player
		//and the FMS application. 
		private var netConnection:NetConnection;
		
		//The NetStream which streams the video, specified in the source property,
		//which will be streamed with the component.  
		private var netStream:NetStream;
		
		//Internal soundTransform to handle volume changing 
		private var _soundTransform:SoundTransform;
		
		//Composited part: Button to stop playback 
		protected var stopButton:Button;
		
		//Composited part: Button to start playback 
		protected var playButton:Button;
		
		//Composited part: Button to pause playback 
		protected var pauseButton:Button;
		
		//Composited part: Muting volume 
		protected var muteButton:Button;
		
		//Composited part: Slider to control playback volume
		protected var volumeTrack:HSlider;
		
		//Holder for the playback controls 
		protected var controlBar:UIComponent;
		
		//Border we customize and add for this component 
		protected var border:IFlexDisplayObject;
		
		//--------------------------------------------------------------------------
    	//
    	//  Properties 
    	//
    	//--------------------------------------------------------------------------
		
		//--------------------------------------------------------------------------
    	//
    	//  autoPlay 
    	//
    	//--------------------------------------------------------------------------
		private var _autoPlay:Boolean = true;
		
		[Bindable("autoPlayChanged")]
	    
	    /**
	     *  Whether or not the video starts playing upon
	     *  startup. 
	     */
	    public function get autoPlay():Boolean
	    {
	        return _autoPlay;
	    }
	    
	    /**
	     *  @private
	     */
	    public function set autoPlay(value:Boolean):void
	    {
	        if (_autoPlay != value)
	        {
	            _autoPlay = value;
	            dispatchEvent(new Event("autoPlayChanged"));
	        }
	    }
	    
	    //----------------------------------
	    //  borderMetrics
	    //----------------------------------
	
	    /**
	     *  Returns an EdgeMetrics object that has four properties:
	     *  <code>left</code>, <code>top</code>, <code>right</code>,
	     *  and <code>bottom</code>.
	     *  The value of each property is equal to the thickness of one side
	     *  of the border around the control, in pixels.
	     */
	    public function get borderMetrics():EdgeMetrics
	    {
	        return (border && border is IRectangularBorder) ?
	                IRectangularBorder(border).borderMetrics : EdgeMetrics.EMPTY;
	    }

		
		//--------------------------------------------------------------------------
    	//
    	//  source 
    	//
    	//--------------------------------------------------------------------------
	    private var _source:String = "";
	    private var sourceChanged:Boolean;
	    [Bindable("sourceChanged")]
	    
	    /**
	     *  Relative path and filename of the FLV file to stream.
	     */
	    public function get source():String
	    {
	        return _source;
	    }
	    
	    /**
	     *  @private
	     */
	    public function set source(value:String):void
	    {
	        if (_source != value)
	        {
	            _source = value;
	            sourceChanged = true;
	            invalidateProperties();
	            dispatchEvent(new Event("sourceChanged"));
	        }
	    }
	    
	    //--------------------------------------------------------------------------
    	//
    	//  volume
    	//
    	//--------------------------------------------------------------------------
	    private var _volume:Number = 4;
	    private var volumeChanged:Boolean;
	    [Bindable("volumeChanged")]
	    
	    /**
	     *  Volume of the video playback, a number between 1 and 10. 
	     *  
	     *  @default 4 
	     */
	    public function get volume():Number
	    {
	        return _volume;
	    }
	    
	    /**
	     *  @private
	     * 
	     */
	    public function set volume(value:Number):void
	    {
	        if (_volume != value)
	        {
	        	//Use dirty flags and storage variables
	        	//to avoid repetitive work. 
	            _volume = value;
	            volumeChanged = true;
	            invalidateProperties();
	            dispatchEvent(new Event("volumeChanged"));
	        }
	    }
	    
	    //--------------------------------------------------------------------------
    	//
    	//  Inherited Methods  
    	//
    	//--------------------------------------------------------------------------
		
		/**
	     *  Our Video component is a composite component, meaning it contains
	     *  subcomponents within itself. createChildren() is where these composited
	     *  subcomponents are created for all subclasses of UIComponent. 
	     *  
	     *  It is important to remember that createChildren() is where static and 
	     *  non data-driven subcomponents should be added. Dynamic, data-driven 
	     *  subcomponents should be added in commitProperties(). Additionally, this 
	     *  is where subcomponent's event listeners should be added. 
	     * 
	     */
		override protected function createChildren():void
    	{
    		//Remember: call super
    		super.createChildren();
    		
    		createBorder();
    		
    		//Always check the existance of parts before
    		//instantiating them anew. Follow the Flex pattern
    		//of instantiating, configuring and attaching. 
    		if (!controlBar)
    			controlBar = new UIComponent();
    		
    		if (!stopButton)
    		{
    			stopButton = new Button();
    			stopButton.label = "stop";
    			stopButton.addEventListener(MouseEvent.CLICK, stopButtonClickHandler);
    			controlBar.addChild(stopButton);
    		}
    		if (!playButton)
    		{
    			playButton = new Button();
    			playButton.label = "play";
    			playButton.addEventListener(MouseEvent.CLICK, playButtonClickHandler);
    			controlBar.addChild(playButton);
    		}
    		if (!pauseButton)
    		{
    			pauseButton = new Button();
    			pauseButton.label = "pause";
    			pauseButton.addEventListener(MouseEvent.CLICK, pauseButtonClickHandler);
    			controlBar.addChild(pauseButton);
    		}
    		if (!muteButton)
    		{
    			muteButton = new Button();
    			muteButton.label = "X";
    			muteButton.toggle = true;
    			muteButton.addEventListener(MouseEvent.CLICK, muteButtonClickHandler);
    			controlBar.addChild(muteButton);
    		}
    		if (!volumeTrack)
    		{
    			volumeTrack = new HSlider();
    			volumeTrack.maximum = 10;
    			volumeTrack.minimum = 0;
    			volumeTrack.tickInterval = 1;
    			volumeTrack.value = _volume;
    			volumeTrack.addEventListener(SliderEvent.CHANGE, volumeSliderChangeHandler);
    			controlBar.addChild(volumeTrack);
    		}
    		addChild(controlBar);
    	}
    	
    	/**
	     *  Manage changes to properties by using a dirty flag 
	     *  to indicate when property values have changed, and then
	     *  committing the cached value (held in a storage variable). 
	     * 
	     */
    	override protected function commitProperties():void
    	{
    		//Remember: always call super
    		super.commitProperties();
    		
    		//Remember: Use flags to filter what work is done
    		if (sourceChanged)
    		{
    			createConnectionAndStream();
    			sourceChanged = false;
    		}
    		if (volumeChanged)
    		{
    			_soundTransform.volume = _volume/10;
    			netStream.soundTransform = _soundTransform;
    			volumeTrack.value = _volume;
    			volumeChanged = false;
    		}
    	}
    	
    	/**
	     *  Handles calculating the sizing logic of our video player. 
	     *  Must set measuredWidth and measuredHeight and optionally, 
	     *  measuredMinWidth and measuredMinHeight. 
	     * 
	     */
    	override protected function measure():void
    	{
    		//Remember: call super()
    		super.measure();
    		
    		//Remember: Bulletproof your code by providing defaults 
    		//for styles if the lookup results in undefined (the style 
    		//was not set)
    		var placement:String = getStyle('controlBarPlacement');
    		if (!placement)
    			placement = "bottom";
    		
    		var bm:EdgeMetrics = borderMetrics;	
				
    		if (placement == "bottom" || placement == "top")
    		{
    			//measuredWidth is the max width of the video or controlbar
    			measuredWidth = Math.max(video.width, controlBar.getExplicitOrMeasuredWidth()) + bm.left;
    			//measuredHeight is the summation of the heights of the video, controlbar 
    			measuredHeight = bm.top + video.height + controlBar.getExplicitOrMeasuredHeight();
    		}
    		else if (placement == "left" || placement == "right")
    		{
    			//measuredWidth is the summation of the widths of the controlbar and video widths 
    			measuredWidth = controlBar.getExplicitOrMeasuredWidth() + video.width;
    			//measuredHeight is the max height of the video or controlbar   
    			measuredHeight = Math.max(controlBar.getExplicitOrMeasuredHeight(), video.height) + bm.top;
    		}

			//Its common to set the minimums to the measured values 
			measuredMinWidth = measuredWidth;
			measuredMinHeight = measuredHeight;
 
    	}
    	
    	/**
	     *  Position and layout the children of the video display. 
	     *  Lots of positioning logic to handle the different values
	     *  of the controlBarPlacement style.  
	     * 
	     */
    	override protected function updateDisplayList(unscaledWidth:Number, unscaledHeight:Number):void
    	{
    		//Remember: call super()
    		super.updateDisplayList(unscaledWidth, unscaledHeight);
    		
    		// Border covers the whole thing.
        	if (border)
        	{
            	border.move(0, 0);
            	border.setActualSize(unscaledWidth, unscaledHeight);
        	}
    		
    		var bm:EdgeMetrics = borderMetrics;
			
	        var x:Number = bm.left;
	        var y:Number = bm.top;
	        
	        // Allocated size.
	        var w:Number = unscaledWidth - bm.left - bm.right;
	        var h:Number = unscaledHeight - bm.top - bm.bottom;
	        
	        //Remember to bulletproof your code by 
	        //providing a default if the style lookup results
	        //in undefined since if its unset. 
	        var placement:String = getStyle('controlBarPlacement');
    		if (!placement)
    			placement = "bottom";
			
			//Size controlbar elements with setActualSize()  
	       	stopButton.setActualSize(50, 20);
	       	playButton.setActualSize(50, 20);
	       	pauseButton.setActualSize(60, 20);
	       	muteButton.setActualSize(35, 20);
	       	volumeTrack.setActualSize(80, 20);
	       	
	       	var tempX:Number = x;
	       	var tempY:Number = y;
	       	
	       	//Position controlbar, video, and controlbar elements  
	       	if (placement == "left")
	       	{
	       		controlBar.setActualSize(90, h);
	       		controlBar.move(x, y);
	       		video.width = w - 90;
	       		video.height = h;
	       		video.x = x + controlBar.width;
	       		video.y = y;
	       		tempY += 2;
	       		playButton.move(x, tempY);
	       		tempY += playButton.height + 2;
	       		pauseButton.move(x, tempY);
	       		tempY += pauseButton.height + 2;
	       		stopButton.move(x,  tempY);
	       		tempY += stopButton.height + 2;
	       		volumeTrack.move(x, tempY);
	       		tempY += volumeTrack.height + 2;
	       		muteButton.move(x, tempY);
	       	}
	       	else if (placement == "right")
	       	{
	       		controlBar.setActualSize(90, h);
	       		controlBar.move(x + w - 90, y);
	       		video.width = w - 90;
	       		video.height = h;
	       		video.x = x;
	       		video.y = y;
	       		tempY += 2;
	       		playButton.move(x, tempY);
	       		tempY += playButton.height + 2;
	       		pauseButton.move(x, tempY);
	       		tempY += pauseButton.height + 2;
	       		stopButton.move(x,  tempY);
	       		tempY += stopButton.height + 2;
	       		volumeTrack.move(x, tempY);
	       		tempY += volumeTrack.height + 2;
	       		muteButton.move(x, tempY);
	       	}
	       	else if (placement == "top")
	       	{
	       		controlBar.setActualSize(w, 30);
	       		controlBar.move(x, y);
	       		video.width = w;
	       		video.height = h - 30;
	       		video.x = x; 
	       		video.y = y + controlBar.height;
	       		tempX += 2;
	       		playButton.move(tempX, y);
	       		tempX += playButton.width + 2;
	       		pauseButton.move(tempX, y);
	       		tempX += pauseButton.width + 4;
	       		stopButton.move(tempX, y);
	       		tempX += stopButton.width + 4;
	       		volumeTrack.move(tempX, y);
	       		tempX += volumeTrack.width + 2;
	       		muteButton.move(tempX, y);
	       	}
	       	else if (placement == "bottom")
	       	{
	       		//Size and position the controlBar
	       		//Since its a UIComponent, setActualSize()
	       		//and move() are the methods to use
	       		controlBar.setActualSize(w, 30);
	       		controlBar.move(x, y + h - 30);
	       		
	       		//Size and position the video
	       		//Since its not a UIComponent, set
	       		//x, y, width and height
	       		video.width = w;
	       		video.height = h - 30;
	       		video.x = x;
	       		video.y = y;
	       		
	       		//Use a temporary variable to keep track
	       		//of positiong logic while positioning
	       		//playback controls. 
	       		tempX += 2;
	       		playButton.move(tempX, y);
	       		tempX += playButton.width + 2;
	       		pauseButton.move(tempX, y);
	       		tempX += pauseButton.width + 4;
	       		stopButton.move(tempX, y);
	       		tempX += stopButton.width + 4;
	       		volumeTrack.move(tempX, y);
	       		tempX += volumeTrack.width + 2;
	       		muteButton.move(tempX, y);
	       	}
		}
	    
	    /**
	     *  Manages style changes 
	     */
	    override public function styleChanged(styleProp:String):void
	    {
	    	super.styleChanged(styleProp);
	    	
	    	//Depending on which style has changed, you'll want to invoke
	    	//the proper invalidation method. Here, controlBarPlacement affects
	    	//the measurement information and drawing of the component. 
	    	if (styleProp == "controlBarPlacement")
	    	{
	    		invalidateSize();
	    		invalidateDisplayList();
	    	}
	    }
	    
	    //--------------------------------------------------------------------------
    	//
    	//  Protected Methods 
    	//
    	//--------------------------------------------------------------------------
	    /**
	     *  Creates the border for the VideoWidget.
	     */
	    protected function createBorder():void
	    {
	        if (!border)
	        {
	            var borderClass:Class = getStyle("borderSkin");
	            if (!borderClass)
	            	borderClass = Class(mx.skins.halo.HaloBorder);
		        
		        border = new borderClass();
		            
	            if (border is ISimpleStyleClient)
	            	ISimpleStyleClient(border).styleName = this;
	            if (border is IUIComponent)
	                IUIComponent(border).enabled = enabled;
	            
	            // Add the border behind all the children.
	            addChildAt(DisplayObject(border), 0);
	
	            invalidateDisplayList();
	        }
	    }
	    
	    //--------------------------------------------------------------------------
    	//
    	//  Public Methods 
    	//
    	//--------------------------------------------------------------------------
			
		/**
		 *  @private
		 */		
		public function play():void
		{
			if (netStream && source)
			{
				netStream.play(source);
			}
		}
		
		/**
		 *  @private
		 */		
		public function pause():void
		{
			if (netStream && source)
			{
				netStream.pause();
			}
		}
		
		/**
		 *  @private
		 */
		public function stop():void
		{
	 		if (netStream)
	 		{
	 			video.clear();
	 			netStream.close();
	 			netConnection.close();
	 		}
		}
		
		//--------------------------------------------------------------------------
    	//
    	//  Helper Methods 
    	//
    	//--------------------------------------------------------------------------
		
		/**
		 *  @private
		 */
		private function createConnectionAndStream():void
		{
			//First close out any open NetConnection or NetStream objects
			//and null out the Flash Player video reference. 
			if (netConnection)
			{
				netConnection.close();
				netConnection = null; 
			}
			if (netStream)
			{
				netStream.close();
				netStream = null;
			}
			if (video)
			{
				video.clear();
				video = null;
			}
			
			//Now create the NetConnection object anew 
			netConnection = new NetConnection();
			netConnection.connect(null);
			
			//Create the NetStream object
			netStream = new NetStream(netConnection);
			_soundTransform = new SoundTransform(_volume/10);
			netStream.soundTransform = _soundTransform;
			netStream.addEventListener(AsyncErrorEvent.ASYNC_ERROR, asyncErrorHandler);
			netStream.addEventListener(NetStatusEvent.NET_STATUS, netStatusEventHandler);
			video = new Video();
			addChild(video);
			video.attachNetStream(netStream);
			
			//Dispatch our ready event. 
			var ve:VideoEvent = new VideoEvent(VideoEvent.READY);
			dispatchEvent(ve);
		}
		
		//--------------------------------------------------------------------------
    	//
    	//  Event handlers
    	//
    	//--------------------------------------------------------------------------
    	
    	/**
	     *  Handles the playing of the video, if autoPlay is set to true,
	     *  upon startup of the component. 
	     * 
	     */
    	private function creationCompleteHandler(event:FlexEvent):void
    	{
    		if (autoPlay)
    		{
    			play();
    		}
    	}
    	
    	/**
	     *  Handles when the user has stopped playback by clicking on the stopButton
	     *  subcomponent. 
	     * 
	     */
    	private function stopButtonClickHandler(event:MouseEvent):void
    	{
    		stop();
    	}
    	
    	/**
	     *  Handles when the user has started playback by clicking on the playButton
	     *  subcomponent. 
	     * 
	     */
    	private function playButtonClickHandler(event:MouseEvent):void
    	{
    		play();
    	}

		/**
	     *  Handles when the user has paused playback by clicking on the pauseButton
	     *  subcomponent. 
	     * 
	     */
    	private function pauseButtonClickHandler(event:MouseEvent):void
    	{
    		pause();
    	}
    	
    	/**
	     *  Handles when the user changes volume by sliding the volume slider.  
	     * 
	     */
    	private function volumeSliderChangeHandler(event:SliderEvent):void
    	{
    		volume = event.value;
    		
    	}
    	
    	/**
	     *  Handles when the user clicks the mute button to mute or unmute   
	     * 
	     */
    	private function muteButtonClickHandler(event:MouseEvent):void
    	{
    		if (muteButton.selected)
    		{
    			//Muted
    			volume = 0;
    		}
    		else
    		{
    			volume = 4;
    			muteButton.selected = false;
    		}
    	}
    	
    	private function netStatusEventHandler(event:NetStatusEvent):void
    	{
    		if (event.info.code == "NetStream.Play.Stop")
    		{
    			//During the demo, I want the video to replay. 
    			//This is just a hack to do this. 
    			netStream.play(source);
    		}
    	}
    	
    	private function asyncErrorHandler(event:AsyncErrorEvent):void
    	{
    		trace(event.text);
    	}
    	
	}	
}	